@@ -26,11 +26,10 @@ from . import validate
 
 import logging
 log = logging.getLogger(__name__)
-log.setLevel(logging.DEBUG)
 log.addHandler(logging.NullHandler())
 
 
-__all__ = ['to_fits', 'from_fits', 'fits_hdu_name', 'get_hdu']
+__all__ = ['to_fits', 'from_fits', 'fits_hdu_name', 'get_hdu', 'is_builtin_fits_keyword']
 
 
 _ASDF_GE_2_6 = parse_version(asdf.__version__) >= parse_version('2.6')
@@ -51,7 +50,7 @@ _builtin_regex = re.compile(
     '|'.join('(^{0}$)'.format(x) for x in _builtin_regexes))
 
 
-def _is_builtin_fits_keyword(key):
+def is_builtin_fits_keyword(key):
     """
     Returns `True` if the given `key` is a built-in FITS keyword, i.e.
     a keyword that is managed by ``astropy.io.fits`` and we wouldn't
@@ -182,7 +181,7 @@ def _get_or_make_hdu(hdulist, hdu_name, index=None, hdu_type=None, value=None):
             new_hdu = _make_hdu(hdulist, hdu_name, index=index,
                                 hdu_type=hdu_type, value=value)
             for key, val in hdu.header.items():
-                if not _is_builtin_fits_keyword(key):
+                if not is_builtin_fits_keyword(key):
                     new_hdu.header[key] = val
             hdulist.remove(hdu)
             hdu = new_hdu
@@ -208,7 +207,7 @@ def _fits_comment_section_handler(validator, properties, instance, schema):
     title = schema.get('title')
     if title is not None:
         current_comment_stack = validator.comment_stack
-        current_comment_stack.append(util.ensure_ascii(title))
+        current_comment_stack.append(ensure_ascii(title))
 
     for property, subschema in properties.items():
         if property in instance:
@@ -238,12 +237,12 @@ def _fits_element_writer(validator, fits_keyword, instance, schema):
         hdu.header.append((' ', ''), end=True)
     validator.comment_stack = []
 
-    comment = util.ensure_ascii(util.get_short_doc(schema))
-    instance = util.ensure_ascii(instance)
+    comment = ensure_ascii(get_short_doc(schema))
+    instance = ensure_ascii(instance)
 
     if fits_keyword in ('COMMENT', 'HISTORY'):
         for item in instance:
-            hdu.header[fits_keyword] = util.ensure_ascii(item)
+            hdu.header[fits_keyword] = ensure_ascii(item)
     elif fits_keyword in hdu.header:
         hdu.header[fits_keyword] = (instance, comment)
     else:
@@ -353,6 +352,25 @@ def _save_from_schema(hdulist, tree, schema):
     validator.validate(tree, _schema=schema)
 
 
+def _normalize_arrays(tree):
+    """
+    Convert arrays in the tree to C-contiguous, since that is
+    how they are written to disk by astropy.io.fits and we
+    don't want the asdf library to notice the change in memory
+    layout and duplicate the array in the embedded ASDF.
+    """
+    def normalize_array(node, json_id):
+        if isinstance(node, np.ndarray):
+            # We can't use np.ascontiguousarray because it converts FITS_rec
+            # to vanilla np.ndarray, which results in misinterpretation of
+            # unsigned int values.
+            if not node.flags.c_contiguous:
+                node = node.copy()
+        return node
+
+    return treeutil.walk_and_modify(tree, normalize_array)
+
+
 def _save_extra_fits(hdulist, tree):
     # Handle _extra_fits
     for hdu_name, parts in tree.get('extra_fits', {}).items():
@@ -364,7 +382,7 @@ def _save_extra_fits(hdulist, tree):
         if 'header' in parts:
             hdu = _get_or_make_hdu(hdulist, hdu_name)
             for key, val, comment in parts['header']:
-                if _is_builtin_fits_keyword(key):
+                if is_builtin_fits_keyword(key):
                     continue
                 hdu.header.append((key, val, comment), end=True)
 
@@ -393,6 +411,7 @@ def to_fits(tree, schema):
     hdulist = fits.HDUList()
     hdulist.append(fits.PrimaryHDU())
 
+    tree = _normalize_arrays(tree)
     _save_from_schema(hdulist, tree, schema)
     _save_extra_fits(hdulist, tree)
     _save_history(hdulist, tree)
@@ -476,14 +495,13 @@ def _load_from_schema(hdulist, schema, tree, context, skip_fits_update=False):
                 hdulist, fits_keyword, schema,
                 ctx.get('hdu_index'), known_keywords)
 
-            if result is None:
-                validate.value_change(path, result, schema,
-                                      context._pass_invalid_values,
-                                      context._strict_validation)
+            if result is None and context._validate_on_assignment:
+                validate.value_change(path, result, schema, context)
             else:
-                if validate.value_change(path, result, schema,
-                                         context._pass_invalid_values,
-                                         context._strict_validation):
+                if context._validate_on_assignment:
+                    if validate.value_change(path, result, schema, context):
+                        properties.put_value(path, result, tree)
+                else:
                     properties.put_value(path, result, tree)
 
         elif 'fits_hdu' in schema and (
@@ -491,14 +509,13 @@ def _load_from_schema(hdulist, schema, tree, context, skip_fits_update=False):
             result = _fits_array_loader(
                 hdulist, schema, ctx.get('hdu_index'), known_datas)
 
-            if result is None:
-                validate.value_change(path, result, schema,
-                                      context._pass_invalid_values,
-                                      context._strict_validation)
+            if result is None and context._validate_on_assignment:
+                validate.value_change(path, result, schema, context)
             else:
-                if validate.value_change(path, result, schema,
-                                         context._pass_invalid_values,
-                                         context._strict_validation):
+                if context._validate_on_assignment:
+                    if validate.value_change(path, result, schema, context):
+                        properties.put_value(path, result, tree)
+                else:
                     properties.put_value(path, result, tree)
 
         if schema.get('type') == 'array':
@@ -526,7 +543,7 @@ def _load_extra_fits(hdulist, known_keywords, known_datas, tree):
 
         cards = []
         for key, val, comment in hdu.header.cards:
-            if not (_is_builtin_fits_keyword(key) or
+            if not (is_builtin_fits_keyword(key) or
                     key in known):
                 cards.append([key, val, comment])
 
@@ -678,13 +695,9 @@ def _verify_skip_fits_update(skip_fits_update, hdulist, asdf_struct, context):
         return False
 
     # Ensure model types match
-    hdulist_class = util._class_from_model_type(hdulist)
-    if hdulist_class is None:
-        log.debug('Cannot determine model of the FITS file.'
-                  ' Cannot skip updating from FITS headers.')
-        return False
-    if not isinstance(context, hdulist_class):
-        log.debug(f'Input model {hdulist_class} does not match the'
+    hdulist_model_type = util.get_model_type(hdulist)
+    if hdulist_model_type != context.__class__.__name__:
+        log.debug(f'Input model type {hdulist_model_type} does not match the'
                   f' requested model {type(context)}.'
                   ' Cannot skip updating from FITS headers.')
         return False
@@ -726,3 +739,24 @@ def fits_hash(hdulist):
             if hdu.name != 'ASDF').encode()
         )
     return fits_hash.hexdigest()
+
+
+def get_short_doc(schema):
+    title = schema.get('title', None)
+    description = schema.get('description', None)
+    if description is None:
+        description = title or ''
+    else:
+        if title is not None:
+            description = title + '\n\n' + description
+    return description.partition('\n')[0]
+
+
+def ensure_ascii(s):
+    # TODO: This function seems to only ever receive
+    # string input.  Also it's not checking that the
+    # characters in the string fall within the valid
+    # range for FITS headers.
+    if isinstance(s, bytes):
+        s = s.decode('ascii')
+    return s
