@@ -1,5 +1,5 @@
 """
-Data model class heirarchy
+Data model class hierarchy
 """
 
 import copy
@@ -16,8 +16,8 @@ from astropy.time import Time
 from astropy.wcs import WCS
 
 import asdf
+from asdf.tags.core import NDArrayType
 from asdf import AsdfFile
-from asdf import yamlutil
 from asdf import schema as asdf_schema
 
 from . import ndmodel
@@ -26,7 +26,7 @@ from . import fits_support
 from . import properties
 from . import schema as mschema
 from . import validate
-from .util import get_envar_as_boolean
+from .util import get_envar_as_boolean, remove_none_from_tree
 from . import s3_utils
 
 from .history import HistoryList
@@ -42,9 +42,6 @@ _DEFAULT_SCHEMA = {
                 "filename": {
                     "type": "string",
                 },
-                "date": {
-                    "type": "string",
-                },
                 "model_type": {
                     "type": "string",
                 },
@@ -63,11 +60,12 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
     """
     The schema URI to validate the model against.  If
     None, only basic validation of required metadata
-    properties (filename, date, model_type) will occur.
+    properties (filename, model_type) will occur.
     """
 
     def __init__(self, init=None, schema=None, memmap=False,
                  pass_invalid_values=None, strict_validation=None,
+                 validate_on_assignment=None,
                  ignore_missing_extensions=True, **kwargs):
         """
         Parameters
@@ -113,6 +111,15 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             If `None`, value will be taken from the environmental STRICT_VALIDATION.
             Otherwise, the default value is `False`.
 
+        validate_on_assignment : bool or None
+            Defaults to 'None'.
+            If `None`, value will be taken from the environmental VALIDATE_ON_ASSIGNMENT,
+            defaulting to 'True' if  no environment variable is set.
+            If 'True', attribute assignments are validated at the time of assignment.
+            Validation errors generate warnings and values will be set to `None`.
+            If 'False', schema validation occurs only once at the time of write.
+            Validation errors generate warnings.
+
         ignore_missing_extensions : bool
             When `False`, raise warnings when a file is read that
             contains metadata about extensions that are not available.
@@ -138,8 +145,12 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
         if strict_validation is None:
             strict_validation = get_envar_as_boolean("STRICT_VALIDATION",
                                                           False)
+        if validate_on_assignment is None:
+            validate_on_assignment = get_envar_as_boolean("VALIDATE_ON_ASSIGNMENT",
+                                                          True)
         self._strict_validation = strict_validation
         self._ignore_missing_extensions = ignore_missing_extensions
+        self._validate_on_assignment = validate_on_assignment
 
         kwargs.update({'ignore_missing_extensions': ignore_missing_extensions})
 
@@ -156,10 +167,14 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
         # Provide the object as context to other classes and functions
         self._ctx = self
 
+        # Initialize with an empty AsdfFile instance as this is needed for
+        # reading in FITS files where validate._check_value() gets called, and
+        # ctx needs to have an _asdf attribute.
+        self._asdf = AsdfFile()
+
         # Determine what kind of input we have (init) and execute the
         # proper code to intiailize the model
-        self._files_to_close = []
-        self._iscopy = False
+        self._file_references = []
         is_array = False
         is_shape = False
         shape = None
@@ -217,7 +232,7 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
                 asdffile = fits_support.from_fits(
                     hdulist, self._schema, self._ctx, **kwargs
                 )
-                self._files_to_close.append(hdulist)
+                self._file_references.append(_FileReference(hdulist))
 
             elif file_type == "asdf":
                 asdffile = self.open_asdf(init=init, **kwargs)
@@ -231,6 +246,8 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             raise ValueError(
                 "Can't initialize datamodel using {0}".format(str(type(init))))
 
+        self._file_references.append(_FileReference(asdffile))
+
         # Initialize object fields as determined from the code above
         self._shape = shape
         self._instance = asdffile.tree
@@ -260,28 +277,6 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             # referenced. Do so now.
             getattr(self, primary_array_name)
 
-        # if the input is from a file, set the filename attribute
-        if isinstance(init, str):
-            self.meta.filename = os.path.basename(init)
-        elif isinstance(init, fits.HDUList):
-            info = init.fileinfo(0)
-            if info is not None:
-                filename = info.get('filename')
-                if filename is not None:
-                    self.meta.filename = os.path.basename(filename)
-
-        # if the input model doesn't have a date set, use the current date/time
-        if not self.meta.hasattr('date'):
-            current_date = Time(datetime.datetime.now())
-            current_date.format = 'isot'
-            self.meta.date = current_date.value
-
-        # store the data model type, if not already set
-        klass = self.__class__.__name__
-        if klass != 'DataModel':
-            if not self.meta.hasattr('model_type'):
-                self.meta.model_type = klass
-
         # initialize arrays from keyword arguments when they are present
 
         for attr, value in kwargs.items():
@@ -291,6 +286,30 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
                 if 'datatype' in subschema:
                     setattr(self, attr, value)
 
+        # Call hook that sets model properties
+        self.on_init(init)
+
+    @property
+    def crds_observatory(self):
+        """
+        Get the CRDS observatory code for this model.
+
+        Returns
+        -------
+        str
+        """
+        raise NotImplementedError("The base DataModel class cannot be used to select best references")
+
+    def get_crds_parameters(self):
+        """
+        Get the parameters used by CRDS to select references for this model.
+
+        Returns
+        -------
+        dict
+        """
+        raise NotImplementedError("The base DataModel class cannot be used to select best references")
+
     @property
     def _model_type(self):
         return self.__class__.__name__
@@ -331,30 +350,15 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.close()
 
-    def _drop_arrays(self):
-        def _drop_array(d):
-            # Walk tree and delete numpy arrays
-            if isinstance(d, dict):
-                for val in d.values():
-                    _drop_array(val)
-            elif isinstance(d, list):
-                for val in d:
-                    _drop_array(val)
-            elif isinstance(d, np.ndarray):
-                del d
-            else:
-                pass
-        _drop_array(self._instance)
-
     def close(self):
-        if not self._iscopy:
-            if self._asdf is not None:
-                self._asdf.close()
-                self._drop_arrays()
-
-            for fd in self._files_to_close:
-                if fd is not None:
-                    fd.close()
+        # This method is called by __del__, which may be invoked
+        # even when the model failed to initialize.  Consequently,
+        # we can't assume that any attributes have been set.
+        if hasattr(self, "_file_references"):
+            for file_reference in self._file_references:
+                file_reference.decrement()
+            # Discard the list in case close is called a second time.
+            self._file_references = []
 
     @staticmethod
     def clone(target, source, deepcopy=False, memo=None):
@@ -362,13 +366,13 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             instance = copy.deepcopy(source._instance, memo=memo)
             target._asdf = AsdfFile(instance)
             target._instance = instance
-            target._iscopy = source._iscopy
         else:
             target._asdf = source._asdf
             target._instance = source._instance
-            target._iscopy = True
+            for file_reference in source._file_references:
+                file_reference.increment()
+                target._file_references.append(file_reference)
 
-        target._files_to_close = []
         target._shape = source._shape
         target._ctx = target
         target._no_asdf_extension = source._no_asdf_extension
@@ -387,35 +391,9 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
 
     def validate(self):
         """
-        Re-validate the model instance againsst its schema
+        Re-validate the model instance against its schema
         """
-        validate.value_change(str(self), self._instance, self._schema,
-                              self._pass_invalid_values,
-                              self._strict_validation)
-
-    def validate_required_fields(self):
-        """
-        Walk the schema and make sure all required fields are
-        in the model
-        """
-        def callback(schema, path, combiner, ctx, recurse):
-            if 'fits_required' not in schema:
-                return
-
-            # Get the value pointed at by the path to the node,
-            # or None in case there is no entry for the node
-
-            node = ctx
-            for attr in path:
-                node = getattr(node, attr)
-                if node is None:
-                    break
-
-            validate.value_change(path, node, schema,
-                                  ctx._pass_invalid_values,
-                                  ctx._strict_validation)
-
-        mschema.walk_schema(self._schema, callback, ctx=self)
+        validate.value_change(str(self), self._instance, self._schema, self)
 
     def info(self, *args, **kwargs):
         return self._asdf.info(**kwargs)
@@ -442,6 +420,31 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             primary_array_name = ''
         return primary_array_name
 
+    def on_init(self, init):
+        """
+        Hook used to customize model attributes at the end of ``__init__``.
+
+        Parameters
+        ----------
+        init : object
+            First argument to ``__init__``.
+        """
+        # if the input is from a file, set the filename attribute
+        if isinstance(init, str):
+            self.meta.filename = os.path.basename(init)
+        elif isinstance(init, fits.HDUList):
+            info = init.fileinfo(0)
+            if info is not None:
+                filename = info.get('filename')
+                if filename is not None:
+                    self.meta.filename = os.path.basename(filename)
+
+        # store the data model type, if not already set
+        klass = self.__class__.__name__
+        if klass != 'DataModel':
+            if not self.meta.hasattr('model_type'):
+                self.meta.model_type = klass
+
     def on_save(self, path=None):
         """
         This is a hook that is called just before saving the file.
@@ -460,13 +463,14 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
         if isinstance(path, str):
             self.meta.filename = os.path.basename(path)
 
-        current_date = Time(datetime.datetime.now())
-        current_date.format = 'isot'
-        self.meta.date = current_date.value
-
         # Enforce model_type to be the actual type of model being saved.
         self.meta.model_type = self._model_type
 
+        # DataModel considers None to be equivalent to missing node, but
+        # asdf 2.8+ writes None as null values, so we need to remove
+        # Nones before proceeding with the save.
+        self._instance = remove_none_from_tree(self._instance)
+
     def save(self, path, dir_path=None, *args, **kwargs):
         """
         Save to either a FITS or ASDF file, depending on the path.
@@ -572,6 +576,7 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             `~asdf.AsdfFile.write_to`.
         """
         self.on_save(init)
+        self.validate()
         asdffile = self.open_asdf(self._instance, **kwargs)
         asdffile.write_to(init, *args, **kwargs)
 
@@ -695,12 +700,6 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             If `return_result` is `True`, a list of the locations in
             the schema where this FITS keyword is used.  Each element
             is a dot-separated path.
-
-        Example
-        -------
-        >>> model = DataModel()
-        >>> model.find_fits_keyword('DATE-OBS')
-        ['meta.observation.date']
         """
         from . import schema
         return schema.find_fits_keyword(self.schema, keyword)
@@ -740,21 +739,6 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
                 raise KeyError(repr(key))
         return meta
 
-    def get_item_as_json_value(self, key):
-        """
-        Equivalent to __getitem__, except returns the value as a JSON
-        basic type, rather than an arbitrary Python type.
-        """
-        assert isinstance(key, str)
-        meta = self
-        parts = key.split('.')
-        for part in parts:
-            try:
-                meta = getattr(meta, part)
-            except AttributeError:
-                raise KeyError(repr(key))
-        return yamlutil.custom_tree_to_tagged_tree(meta, self._instance)
-
     def __setitem__(self, key, value):
         """
         Set a metadata value using a dotted name.
@@ -781,7 +765,7 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
         else:
             meta[part] = value
 
-    def iteritems(self):
+    def items(self):
         """
         Iterates over all of the schema items in a flat way.
 
@@ -806,10 +790,7 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
         for x in recurse(self._instance):
             yield x
 
-    # We are just going to define the items to return the iteritems
-    items = iteritems
-
-    def iterkeys(self):
+    def keys(self):
         """
         Iterates over all of the schema keys in a flat way.
 
@@ -818,20 +799,16 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
         `meta.observation.date` will end up in the result as the
         string `"meta.observation.date"`.
         """
-        for key, val in self.iteritems():
+        for key, val in self.items():
             yield key
 
-    keys = iterkeys
-
-    def itervalues(self):
+    def values(self):
         """
         Iterates over all of the schema values in a flat way.
         """
-        for key, val in self.iteritems():
+        for key, val in self.items():
             yield val
 
-    values = itervalues
-
     def update(self, d, only=None, extra_fits=False):
         """
         Updates this model with the metadata elements from another model.
@@ -972,10 +949,10 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
             return val
 
         if include_arrays:
-            return dict((key, convert_val(val)) for (key, val) in self.iteritems())
+            return dict((key, convert_val(val)) for (key, val) in self.items())
         else:
-            return dict((key, convert_val(val)) for (key, val) in self.iteritems()
-                        if not isinstance(val, np.ndarray))
+            return dict((key, convert_val(val)) for (key, val) in self.items()
+                        if not isinstance(val, (np.ndarray, NDArrayType)))
 
     @property
     def schema(self):
@@ -984,15 +961,6 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
     def get_fileext(self):
         return 'fits'
 
-    # TODO: This is just here for backward compatibility
-    @property
-    def _extra_fits(self):
-        return self.extra_fits
-
-    # TODO: For backward compatibility
-    def get_section(self, name):
-        return getattr(self, name)
-
     @property
     def history(self):
         """
@@ -1119,3 +1087,27 @@ class DataModel(properties.ObjectNode, ndmodel.NDModel):
         if attribute in self.instance:
             return getattr(self, attribute)
         raise AttributeError(f'{self} has no attribute "{attribute}"')
+
+
+class _FileReference:
+    """
+    Reference counter for open file pointers managed by
+    DataModel.  Once decremented to zero the file will
+    be closed.
+    """
+
+    def __init__(self, file):
+        self._file = file
+        self._count = 1
+
+    def increment(self):
+        self._count += 1
+
+    def decrement(self):
+        if self._count <= 0:
+            return
+
+        self._count -= 1
+        if self._count <= 0:
+            self._file.close()
+            self._file = None
